"""
minEER utilites
"""
from typing import List
from .mineer import minEER
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
import numpy as np
import pandas as pd
import os, functools, random

default_nreads = 5000
default_mal = 100
default_mae = 1e-2

class File:
    """Fastq file"""
    def __init__(self, filepath: str, direction: str, mal: int, mae: float, suffix: str=None):
        self.filepath = filepath
        assert direction in ['f', 'r'], 'Direction must be "f" or "r"'
        self.direction = direction
        self.suffix = suffix
        self.sample = os.path.basename(filepath).split(suffix)[0]
        self.mal = mal
        self.mae = mae

    @functools.cached_property
    def reads(self):
        """Load all reads"""
        return [Read(r, self.mal, self.mae, self) for r in SeqIO.parse(self.filepath, 'fastq')]


class Record:
    """A single read"""
    def __init__(self, seqrecord):
        self.record = seqrecord
        
    @functools.cached_property
    def length(self):
        """Read length"""
        return len(self.record)

    @functools.cached_property
    def phred(self):
        """Extract phred quality scores"""
        return self.record.letter_annotations['phred_quality']
    
    @functools.cached_property
    def ee(self):
        """Get expected error scores"""
        return phred2ee(self.phred)
    
class Read:
    """Untrimmed and Trimmed data on read"""
    def __init__(self, seqrecord: SeqRecord, mal: int, mae: float, file: File=None):
        self.untrimmed = Record(seqrecord)
        self.id = seqrecord.id
        self.file = file
        self.mal = mal
        self.mae = mae

        # minEER methods
        self.mineer = False
        self.pass_qc_mineer = None
        self.trimpos_mineer = None

        # Truncation methods
        self.trimmed = None
        self.pass_e = None
        self.pass_l = None
        self.pass_qc = None
        self.bothpassing = None # Updated from ReadPair

    def runMineer(self):
        """Run minEER on Read"""
        # Run minEER to get truncation positions
        trimpos_mineer = minEER(self.untrimmed.ee, self.mal, self.mae)
        self.trimpos_mineer = trimpos_mineer
        self.mineer = True
        self.pass_qc_mineer = not trimpos_mineer[0] is None

    def truncate(self, trimpos):
        """Truncate sequence given """
        trimstart, trimend = trimpos
        trimlen = trimend - trimstart
        # Truncate
        trimmed_seqrecord = self.untrimmed.record[trimstart: trimend]
        self.trimmed = Record(trimmed_seqrecord)
        # Passes QC?
        self.pass_e = self.trimmed.ee.mean() <= self.mae
        self.pass_l = self.trimmed.length == trimlen
        self.pass_qc = self.pass_e & self.pass_l


class ReadPair:
    """A forward/rev read pair"""
    def __init__(self, fwd_read: Read, rev_read:Read = None, filter: str='both'):
        self.fwd_read = fwd_read
        self.rev_read = rev_read
        self.paired = bool(rev_read)
        self.filter = filter

    def truncBoth(self, fwd_pos: tuple, rev_pos: tuple=None):
        """Truncate both read pairs"""
        self.fwd_read.truncate(fwd_pos)
        if self.paired:
            self.rev_read.truncate(rev_pos)
    
    @property
    def bothpassing(self):
        """Check that both reads are passing qc"""
        # Check that read has been truncated, don't evaluate qc if not
        if not self.fwd_read.trimmed:
            return

        reads = [self.fwd_read]
        if self.paired:
            reads.append(self.rev_read)
        
        # Determine passing using `filter`
        # Filter out readpairs where any read fails. Keep == all passing
        if self.filter == 'any': 
            bp = all([r.pass_qc for r in reads])
        # Filter only when both pass. Keep == any passing 
        if self.filter == 'both':
            bp = any([r.pass_qc for r in reads])
        # Don't filter any based on quality. Keep == all passing length
        if self.filter == 'no':
            bp = all([r.pass_l for r in reads])

        # Update method for each Read and for ReadPair
        self.fwd_read.bothpassing = bp
        if self.paired:
            self.rev_read.bothpassing = bp
        return bp

class Sample:
    """Readpairs for all reads in both files of a sample (for paired reads). Assumes reads are in order for each pair"""
    def __init__(self, name: str, fwd_file: File, rev_file: File=None, filter: str='both'):
        self.name = name
        self.fwd_file = fwd_file
        self.readpairs = None
        if rev_file:
            self.rev_file = rev_file
            self.readpairs = [ReadPair(f, r, filter) for f, r in zip(self.fwd_file.reads, self.rev_file.reads)]
        else:
            self.readpairs = [ReadPair(f, filter=filter) for f in self.fwd_file.reads]
    

def phred2ee(phred):
    """
    Convert a Phred score to an expected error probability
    ee = 10 ^ (-P/10)
    """
    phred = np.array(phred)
    ee = np.power(10, -phred / 10)
    return ee

class Project:
    """
    Collects samples for project

    Inputs:
    | filepaths: Array of file path locations. Files without a file suffix will be ignored
    | {fwd, rev}_format: Suffix of forward and reverse files
        ex. For forward read A_1.fastq and reverse read A_2.fastq
        fwd_format = '_1.fastq'
        rev_format = '_2.fastq'
        * If single ended, do not define rev_format
    | nreads: Number of reads to sample for computing truncation positions (default: 10000)
    | mal: Maximum acceptable length (default: 100)
    | mae: Maximum acceptable error (default: 1e-2)
    | aggmethod: Method to aggregate truncation positions across reads, either 'median' (default) or 'mean'
    | filter: 
        filter='any':  Filter out all readpairs where any read where EER > mae
        filter='both': Only filter out readpairs where both reads EER > mae [Default]
        filter='no':   Do not filter out any reads based on EER
        * All reads with length < truncation length are always filtered (len(all reads) == truncation length)
    | outdir: Output directory
    | no_shuffle: Don't shuffle reads before sampling (just for testing)

    Pipeline structure
    * Project: Collection of samples
    * Sample: Collection of read pairs
    * ReadPair: Collection of pair of reads
    * Read: Untrimmed and trimmed record
    * Record: A single SeqRecord with extracted data
    """
    def __init__(self, filepaths: List[str], fwd_format: str, rev_format: str=None, nreads: int=default_nreads, mal: int=default_mal, mae: float=default_mae, aggmethod: str='median', filter: bool='both', outdir: str=None, no_shuffle: bool=False):
        #TODO: MAKE SURE ABSPATH WORKS
        self.paired = bool(rev_format)
        assert fwd_format != rev_format, 'If "rev_format" is provided, it must differ from "fwd_format".\nOnly provide "fwd_format" for single end mode.'
        self.fwd_format = fwd_format
        self.rev_format = rev_format
        suffixes = tuple([fmt for fmt in [fwd_format, rev_format] if fmt])
        self.filepaths = sorted([f for f in filepaths if f.endswith(suffixes)])
        self.nreads = nreads
        self.mal = mal
        self.mae = mae
        assert aggmethod in ['mean', 'median'], '"aggmethod" must be either "median" (default) or "mean"'
        self.aggmethod = aggmethod
        self.aggfunc = np.median if aggmethod == 'median' else np.mean
        assert filter in ['any', 'both', 'no'], '"filter" must be either "any", "both", or "no"'
        self.filter = filter
        if not outdir:
            outdir = os.getcwd()
        self.outdir = outdir
        self.no_shuffle = no_shuffle

        # All reads
        self.fwd_reads: List[Read] = []
        self.rev_reads : List[Read]= []
        self.samples: List[Sample] = None
        self._sample_mapping = None
        # Subset for minEER
        self.fwd_sub: List[Read] = None
        self.fwd_n_tries = None
        self.fwd_frac_passing = None
        self.rev_sub: List[Read] = None
        self.rev_n_tries = None
        self.rev_frac_passing = None
        self.all_sub = []
        # Global
        self.fwd_pos = None
        self.fwd_len = None
        self.rev_pos = None
        self.rev_len = None
        self.passing_readpairs: List[ReadPair] = None

    @functools.cached_property
    def files(self) -> List[File]:
        """All files in project"""
        fs = []
        for path in self.filepaths:
            direction = 'f' if path.endswith(self.fwd_format) else 'r'
            suffix = self.fwd_format if direction == 'f' else self.rev_format
            file = File(path, direction, self.mal, self.mae, suffix)
            fs.append(file)
        return fs

    def getReadsandSamples(self):
        """Sort reads into forward and reverse reads and create samples"""
        samples = {}
        for file in self.files:
            # Create sample pairs
            sample = file.sample
            direction = file.direction
            if sample not in samples:
                samples[sample] = {'f': None, 'r': None}
            samples[sample][direction] = file

            # Sort reads by direction
            reads = file.reads
            if direction == 'f':
                attr = 'fwd_reads'
            else:
                attr = 'rev_reads'
            for read in reads:
                # Save read to direction
                getattr(self, attr).append(read)

        # Create sample objects
        Samples = [Sample(sample, files['f'], files['r'], self.filter) for sample, files in samples.items()]
        self.samples = Samples
        self._sample_mapping = samples

    @property
    def _reportInputs(self):
        """Report number of samples and reads"""
        if not self.samples:
            self.getReadsandSamples()
        # Basic inputs
        args = ['paired', 'fwd_format', 'rev_format', 'nreads', 'mal', 'mae', 'aggmethod', 'filter', 'outdir']
        nfiles = len(self.files)
        nsamples = len(self.samples)
        nreadpairs = len([rp for s in self.samples for rp in s.readpairs])
        pairs = {arg: getattr(self, arg) for arg in args}
        pairs.update({'Files': nfiles, 'Samples': nsamples, 'Read Pairs': nreadpairs})

        order = args + ['Files', 'Samples', 'Read Pairs']
        input_report = alignedSpacing(pairs, 50, order)

        # Sample-pairs
        sample_map_data = []
        for s, file_dict in self._sample_mapping.items():
            s_data = {'Sample': s, 'Forward': os.path.basename(file_dict['f'].filepath)}
            if self.paired:
                s_data.update({'Reverse': os.path.basename(file_dict['r'].filepath)})
            sample_map_data.append(s_data)
        sample_map_df = pd.DataFrame(sample_map_data).set_index('Sample')

        return input_report, sample_map_df

    def subsampleReads(self, reads: List[Read]):
        """Subsample `nreads` and keep those that pass minEER"""
        if not self.no_shuffle: # ONLY EVER TRUE FOR TESTING
            # Shuffle reads
            random.shuffle(reads)

        # Keep track of passing
        read_count = 0
        passing_reads = []

        # Get passing reads until `nreads` is reached
        for read in reads:
            read_count += 1
            read.runMineer()
            if read.pass_qc_mineer:
                passing_reads.append(read)
            
            # Break if `nreads` has been reached
            if len(passing_reads) == self.nreads:
                break
        return read_count, passing_reads
    
    def subsampleAll(self):
        """Subsample all reads up to `nreads` keeping those that pass minEER and recording fraction that pass"""
        self.fwd_n_tries, self.fwd_sub = self.subsampleReads(self.fwd_reads)
        self.fwd_frac_passing = len(self.fwd_sub) / self.fwd_n_tries
        self.all_sub.extend(self.fwd_sub)
        if self.paired:
            self.rev_n_tries, self.rev_sub = self.subsampleReads(self.rev_reads)
            self.rev_frac_passing = len(self.fwd_sub) / self.rev_n_tries
            self.all_sub.extend(self.rev_sub)
    
    @property
    def _reportMineerStats(self):
        """Report fraction of passing reads from minEER run"""
        passing_stats = f'From subset run on minEER:\n\tForward reads passing: {len(self.fwd_sub)}/{self.fwd_n_tries} ({self.fwd_frac_passing * 100:.2f})%\n'
        if self.paired:
            passing_stats += f'\tReverse reads passing: {len(self.rev_sub)}/{self.rev_n_tries} ({self.rev_frac_passing * 100:.2f})%\n'

        """Report stats on trimpositions from minEER run"""
        fwd_poses = [{'trimstart': r.trimpos_mineer[0], 'trimend': r.trimpos_mineer[1], 'direction': 'f'} for r in self.fwd_sub if r.pass_qc_mineer]
        data = fwd_poses
        if self.paired:
            rev_poses = [{'trimstart': r.trimpos_mineer[0], 'trimend': r.trimpos_mineer[1], 'direction': 'r'} for r in self.rev_sub if r.pass_qc_mineer]
            data.extend(rev_poses)
        trimpos_df = pd.DataFrame(data)
        trimpos_stats = trimpos_df.groupby('direction').agg(['min', 'max', 'mean', 'median'])
        return passing_stats, trimpos_stats

    def calcPos(self):
        """Calculate global truncation positions from subset"""
        # Get forward trim positions
        fwd_positions = [r.trimpos_mineer for r in self.fwd_sub if r.pass_qc_mineer]
        self.fwd_passing_sub = len(fwd_positions)/self.nreads
        self.fwd_pos = self.aggfunc(fwd_positions, 0).astype(int)
        self.fwd_len = self.fwd_pos[1] - self.fwd_pos[0]
        if self.paired:
            rev_positions = [r.trimpos_mineer for r in self.rev_sub if r.pass_qc_mineer]
            self.rev_passing_sub = len(rev_positions)/self.nreads
            self.rev_pos = self.aggfunc(rev_positions, 0).astype(int)
            self.rev_len = self.rev_pos[1] - self.rev_pos[0]

    @property
    def _reportGlobalPos(self):
        """Report global truncation positions"""
        text = f"""Using {self.aggmethod} and {self.nreads} reads, truncation positions are:
        Forward: {self.fwd_pos}
        """
        if self.paired:
            text += f'Reverse: {self.rev_pos}'
        return text

    def truncAndFilter(self):
        """Truncate all reads to global positions and indicate passing ReadPairs"""
        assert not self.fwd_pos is None, 'Run `calcPos()` for global truncation parameters (or set `fwd_pos` and `rev_pos` manually)'

        passing_rps = []
        for sample in self.samples:
            readpairs = sample.readpairs
            for rp in readpairs:
                rp.truncBoth(self.fwd_pos, self.rev_pos)
                # Save read pairs that both pass QC
                if rp.bothpassing:
                    passing_rps.append(rp)
        
        self.passing_readpairs = passing_rps

    @property
    def _reportTruncStats(self):
        """Report passing reads after using global truncation positions"""
        report_dict = {
            'Forward reads': f'{len([r for r in self.fwd_reads if r.pass_qc])}/{len(self.fwd_reads)}',
            'Reverse reads': f'{len([r for r in self.rev_reads if r.pass_qc])}/{len(self.rev_reads)}',
            'Read pairs': f'{len(self.passing_readpairs)}/{len([rp for s in self.samples for rp in s.readpairs])}'
        }
        if self.paired:
            order = ['Forward reads', 'Reverse reads', 'Read pairs']
        else:
            order = ['Forward reads', 'Read pairs']

        report = 'Reads that pass quality control after truncation:\n'
        report += alignedSpacing(report_dict, order=order)
        return report

    def writeFile(self, file: File):
        """Write one file with truncated sequences. Use File to preserve original order"""
        # Generate output path
        outpath = os.path.join(self.outdir, f'{file.sample}_mineer{file.suffix}')
        # Create output directory if it doesn't exist
        if not os.path.isdir(self.outdir):
            os.makedirs(self.outdir)
        # Gather truncated reads where pairs pass qc
        truncated = [r.trimmed.record for r in file.reads if r.bothpassing]
        # Save
        SeqIO.write(truncated, outpath, 'fastq')
            
    def writeFiles(self):
        """Write all truncated files"""
        for f in self.files:
            self.writeFile(f)

def alignedSpacing(pairs, width = 30, order=None):
    """Generate text with aligned spacing for each element of `pairs` of `width`"""
    text = ''
    if not order:
        order = pairs.keys()
    for k in order:
        v = str(pairs[k])
        spaces = ' ' * (width - len(k) - len(v))
        line = f'{k}:{spaces}{v}\n'
        text += line
    return text


"""
Test mineer utils
"""
from mineer.utils import Project, ReadPair, Read
from Bio.SeqRecord import SeqRecord
from unittest import TestCase
from copy import copy

class TestProjectPaired(TestCase):
    def setUp(self):
        samples = ['a', 'b']
        suffixes = ['_1.fastq', '_2.fastq']
        filepaths = [f'{sample}{suffix}' for sample in samples for suffix in suffixes]
        self.kwargs = dict(
            filepaths = filepaths,
            fwd_format = suffixes[0],
            rev_format = suffixes[1],
            nreads = 10,
            mal = 20,
            mae = 1,
            aggmethod = 'mean',
            outdir = 'test'
        )
        self.project = Project(**self.kwargs)

    def test_project_ingest(self):
        for arg, val in self.kwargs.items():
            project_val = getattr(self.project, arg)
            self.assertEqual(val, project_val)
    
    def test_ignore_bad_files(self):
        bad_paths = ['bad', 'paths']
        expected = copy(self.kwargs['filepaths'])
        self.kwargs['filepaths'].extend(bad_paths)
        self.assertListEqual(self.project.filepaths, expected)

class TestProjectSingle(TestCase):
    def setUp(self):
        filepaths = ['a.fastq', 'b.fastq']
        self.kwargs = dict(
        filepaths = filepaths,
        fwd_format = '.fastq',
        nreads = 10,
        mal = 20,
        mae = 1,
        aggmethod = 'mean',
        outdir = 'test'
        )
        self.project = Project(**self.kwargs)

    def test_project_ingest(self):
        for arg, val in self.kwargs.items():
            project_val = getattr(self.project, arg)
            self.assertEqual(val, project_val)
    
    def test_not_paired(self):
        self.assertFalse(self.project.paired)

    def test_only_forward(self):
        bad_paths = ['a_2.fastq', 'b_2.fastq']
        expected = copy(self.kwargs['filepaths'])
        self.kwargs['filepaths'].extend(bad_paths)
        self.assertListEqual(self.project.filepaths, expected)

class FakeRead:
    # Hack the system
    def __init__(self, pass_l, pass_e):
        self.trimmed = True
        self.pass_l = pass_l
        self.pass_e = pass_e
        self.pass_qc = self.pass_e & self.pass_l

class TestReadPair(TestCase):
    def test_filter_any(self):
        # Bothpassing case
        r1 = FakeRead(True, True)
        r2 = FakeRead(True, True)
        rp = ReadPair(r1, r2, filter='any')
        self.assertTrue(rp.bothpassing)

        # Failing
        r2 = FakeRead(True, False)
        rp = ReadPair(r1, r2, 'any')
        self.assertFalse(rp.bothpassing)
    
    def test_filter_both(self):
        # Bothpassing case
        r1 = FakeRead(True, True)
        r2 = FakeRead(True, False)
        rp = ReadPair(r1, r2, filter='both')
        self.assertTrue(rp.bothpassing)

        # Failing
        r1 = FakeRead(True, False)
        rp = ReadPair(r1, r2, 'both')
        self.assertFalse(rp.bothpassing)

    def test_filter_no(self):
        # Bothpassing case
        r1 = FakeRead(True, False)
        r2 = FakeRead(True, False)
        rp = ReadPair(r1, r2, filter='no')
        self.assertTrue(rp.bothpassing)

        # Failing
        r1 = FakeRead(False, True)
        r2 = FakeRead(False, False)
        rp = ReadPair(r1, r2, 'no')
        self.assertFalse(rp.bothpassing)     


"""
minEER pipeline

Github: https://github.com/michaelsilverstein/minEER
AUTHOR: MICHAEL SILVERSTEIN
EMAIL: michael.silverstein4@gmail.com
"""
import argparse
import random, os
from .utils import Project, default_nreads, default_mal, default_mae
from .viz import Viz
from argparse import ArgumentParser, RawTextHelpFormatter
import pandas as pd


def truncPipeline(filepaths: list, fwd_format: str, rev_format = None, mal=default_mal, mae=default_mae, aggmethod: str='median', nreads=default_nreads, filter='both', outdir=None, viz_outdir=None, no_shuffle=False, write=True):
    """
    Leave `rev_format` blank for single read mode

    See mineer.utils.Project for documentation on each argument

    Run minEER truncation pipeline
    1) Ingest files
    2) Run minEER on subset
    3) Determine global truncation positions
    4) Truncate all reads to global positions and filter out read pairs that don't pass QC
    5) Save truncated sequences
    6) Produce viz if directory provided
    """
    print('******** STARTING MINEER PIPELINE ********')
    # Create project
    project = Project(filepaths, fwd_format, rev_format, nreads, mal, mae, aggmethod, filter, outdir, no_shuffle)
    # 1) Ingest reads
    project.getReadsandSamples()
    # Print inputs
    print('\t\t**** INPUTS ****')
    input_report, sample_map_df = project._reportInputs
    print(input_report)
    print('SAMPLE-FILE PAIRS')
    with pd.option_context('display.max_rows', None):
        print(sample_map_df)
    print()

    # 2) Subsample
    print('\t\t**** RUNNING MINEER ALGORITHM ****')
    print('Running minEER on a subset of %d reads per direction...' % nreads)
    project.subsampleAll()
    print('Complete.')
    print()
    # Print truncstats
    print('\t\t**** minEER truncation stats from subset ****')
    passing_stats, mineer_stats = project._reportMineerStats
    print(passing_stats)
    print(mineer_stats.T)
    print()

    # 3) Determine global truncation positions
    project.calcPos()
    print('\t\t**** GLOBAL TRUNCATION POSITIONS ****')
    globpos_report = project._reportGlobalPos
    print(globpos_report)
    print()

    # 4) Truncate all reads to global positions and filter out read pairs that don't pass QC
    print('\t\t**** TRUNCATING ALL FILES TO GLOBAL POSITIONS ****')
    project.truncAndFilter()
    truncstats = project._reportTruncStats
    print(truncstats)
    print()

    # 5) Save truncated sequences
    if write: # If working with truncPipeline() directly, may not want to write out files
        print('\t\t**** SAVING TRUNCATED FILES ****')
        print(f'Saving files to "{project.outdir}" with format <sample>_mineer<suffix>')
        project.writeFiles()

    # 6) Produce viz if directory provided
    if viz_outdir:
        print('\t\t**** GENERATING FIGURES ****')
        v = Viz(project, viz_outdir)
        v.genFigs()
        print(f'Figures have been saved to "{viz_outdir}"')
    print('******** MINEER RUN COMPLETE ********')

    return project

def mineer_cli(args=None):
    """Command line interface for running minEER"""
    parser = ArgumentParser(formatter_class=RawTextHelpFormatter, description=__doc__)
    parser.add_argument('-i', help='Path to directory containing all (unzipped!) fastq files for a single project', required=True, dest='filepaths')
    parser.add_argument('-f', help="""Forward read filename suffix for paired fastqs or single read filename suffix
        Ex. sample1_1.fastq
            -f _1.fastq
        If extension contains a '-', like sample1-r1.fastq:
            -f="-r1.fastq"
    """, required=True, dest='fwd_format')
    parser.add_argument('-r', help='Reverse read filename suffix (leave blank for single end)', dest='rev_format')
    parser.add_argument('--mal', help='Minimum acceptable length. Default: %d' % default_mal, type=int, default=default_mal)
    parser.add_argument('--mae', help='Maximum acceptable error. Deafult: %f' % default_mae, type=float, default=default_mae)
    parser.add_argument('-m', help='Aggregation method for computing truncation. Default: "median"', choices=['mean', 'median'], default='median', dest='aggmethod')
    parser.add_argument('-n', help='Number of reads to subsample per direction for computing truncation position. Default: %d' % default_nreads, type=int, default=default_nreads, dest='nreads')
    parser.add_argument('--filter', help="""How to filter out truncated reads
        --filter any: Filter out read pairs where any read has EER > mae
        --filter both [Default]: Only filter read pairs where both reads have EER > mae
        --filter no: Do not filter read pairs based on EER
        * In all cases, reads that fall within truncation positions will be filtered""", choices=['any', 'both', 'no'], default='both')
    parser.add_argument('-o', help='Output directory. Default: current working directory', default=os.getcwd(), dest='outdir')
    parser.add_argument('-v', help='Provide output directory to generate and visualizations', dest='viz_outdir')
    parser.add_argument('--test', help=argparse.SUPPRESS, action='store_true', default=False, dest='no_shuffle')

    args = parser.parse_args(args)
    
    # Get filepaths
    filepaths = [os.path.abspath(os.path.join(args.filepaths, f)) for f in os.listdir(args.filepaths)]

    # Extract inputs
    inputs = args.__dict__
    inputs['filepaths'] = filepaths

    # Run pipeline
    project = truncPipeline(**inputs)

    return project

def run():
    # Seperate function to access `project` when testing, but avoid returning object to command line
    mineer_cli()


"""
Download test files
"""

import os, requests

def download(outdir=None):
    # URLs to test files
    root = 'https://raw.githubusercontent.com/michaelsilverstein/database/main/mineer/'
    files = ['SRR9660307_1.fastq', 'SRR9660307_2.fastq', 'SRR9660321_1.fastq', 'SRR9660321_2.fastq']
    urls = list(map(lambda f: os.path.join(root, f), files))

    if not outdir:
        outdir = 'sample_files'
    
    if not os.path.isdir(outdir):
        os.makedirs(outdir)
        
    for url in urls:
        filename = os.path.basename(url)
        outpath = os.path.join(outdir, filename)
        print(f'Downloading {filename} to {outpath}')
        with requests.get(url) as resp:
            with open(outpath, 'wb') as fh:
                fh.write(resp.content)


import setuptools
from mineer import __version__

setuptools.setup(
    name='mineer',
    version=__version__,
    entry_points={'console_scripts':[
        'mineer=mineer.pipeline:run',
        'mineer-test-files=mineer.download_test_files:download'
    ]},
    install_requires= ['numpy', 'pandas', 'biopython', 'seaborn', 'numba'],
    url='',
    license='MIT',
    author='michaelsilverstein',
    author_email='michael.silverstein4@gmail.com',
    description='Automated sequence truncation algorithm',
    packages=['mineer'],
    long_description_content_type = "text/markdown",
    python_requires='>3.8'
)



"""
minEER - A sequence trimming algorithm that seeks to minimize the expected error rate and maximize sequence length

AUTHOR: MICHAEL SILVERSTEIN
EMAIL: michael.silverstein4@gmail.com
"""
import numpy as np
from numba import jit

@jit(nopython=True)
def eerspace(ee, mal):
    """
    Compute expected error rate space where each row is starting position and each column is an ending position

    For each starting position, calculate the expected error rate of every subsequence of the minimum acceptable length
    """
    # Initialize matrix
    n_pos = len(ee)
    eer_space = np.empty((n_pos, n_pos))
    eer_space[:] = np.nan
    # Compute the cumulative sum of the expected errors
    cumsum_ee = np.cumsum(ee)
    # Get normalizing vector to compute EER at each position
    norm = np.arange(n_pos) + 1
    # Initialize first start
    eer_space[0, mal:] = (cumsum_ee / norm)[mal:]

    # For each starting position
    for start in range(1, n_pos - mal):
        # Get cumulative sum from this start position
        cumsum_ee_start = cumsum_ee[start:]

        # Calculate the expected errors from this start position
        eers = (cumsum_ee_start - cumsum_ee[start - 1]) / norm[:-start]
        # Only take the expected errors starting at the minimum acceptable length
        eers_mal = eers[mal:]
        # Update the search space
        eer_space[start, (start + mal) :] = eers_mal
    return eer_space


def minEER(ee, mal=100, mae=1e-3):
    """
    Exhaustive expected error rate (EER) search for a sequence
    Finds the longest sequence with an EER<=thresh with length>=mal
    Inputs:
    | ee <array>: Vector of expected errors
    | mal <int>: Minimum acceptable length of sequence
    | mae <float>: Minimum acceptable EER threshold
    Outputs:
    | {start, end}_pos: Positions where subsequence starts and ends
    """
    # Ensure np.array()
    ee = np.array(ee)
    # Calculate expected error space
    eer_space = eerspace(ee, mal)

    """Find max length sequence below threshold"""
    # Filter out nans to avoid warning about trying to evaluate np.nan<thresh
    idxs = np.array(np.where(np.isfinite(eer_space)))
    threshed = eer_space[idxs[0], idxs[1]] <= mae
    # If no subsequences, return None
    if not threshed.any():
        start_pos, end_pos = None, None
    else:
        # Start and end indices that satisfy threshold
        start_ids, end_ids = idxs[:, threshed]
        # Find longest subsequence
        max_idx = np.argmax(end_ids - start_ids)
        start_pos, end_pos = start_ids[max_idx], end_ids[max_idx]
        # Add one to end pos to be inclusive when indexing
        end_pos += 1
    return start_pos, end_pos 